//
//  Copyright (c) Samuel Défago. All rights reserved.
//
//  License information is available from the LICENSE file.
//

#import "TableViewCellsDemoViewController.h"

#import "FooterView.h"
#import "HeaderView.h"
#import "ProgrammaticTableViewCell.h"
#import "XibTableViewCell.h"

// Categories
typedef NS_ENUM(NSInteger, CellCategoryIndex) {
    CellCategoryIndexEnumBegin = 0,
    CellCategoryIndexSimple = CellCategoryIndexEnumBegin,
    CellCategoryIndexCustom,
    CellCategoryIndexEnumEnd,
    CellCategoryIndexEnumSize = CellCategoryIndexEnumEnd - CellCategoryIndexEnumBegin
};

// Simple cells
typedef NS_ENUM(NSInteger, SimpleCellIndex) {
    SimpleCellIndexEnumBegin = 0,
    SimpleCellIndexDefault = SimpleCellIndexEnumBegin,
    SimpleCellIndexValue1,
    SimpleCellIndexValue2,
    SimpleCellIndexSubtitle,
    SimpleCellIndexEnumEnd,
    SimpleCellIndexEnumSize = SimpleCellIndexEnumEnd - SimpleCellIndexEnumBegin
};

// Custom cells
typedef NS_ENUM(NSInteger, CustomCellIndex) {
    CustomCellIndexEnumBegin = 0,
    CustomCellIndexXib = CustomCellIndexEnumBegin,
    CustomCellIndexProgrammatically,
    CustomCellIndexEnumEnd,
    CustomCellIndexEnumSize = CustomCellIndexEnumEnd - CustomCellIndexEnumBegin
};

@interface TableViewCellsDemoViewController ()

@property (nonatomic, weak) IBOutlet UITableView *tableView;

@end

@implementation TableViewCellsDemoViewController

#pragma mark View lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.tableView.sectionHeaderHeight = [HeaderView height];
    self.tableView.sectionFooterHeight = [FooterView height];
    
    self.tableView.dataSource = self;
    self.tableView.delegate = self;
}

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
    [self.tableView deselectRowAtIndexPath:[self.tableView indexPathForSelectedRow] animated:animated];
}

#pragma mark UITableViewDataSource protocol implementation

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
    return CellCategoryIndexEnumSize;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    switch (section) {
        case CellCategoryIndexSimple: {
            return SimpleCellIndexEnumSize;
            break;
        }
            
        case CellCategoryIndexCustom: {
            return CustomCellIndexEnumSize;
            break;
        }
            
        default: {
            return 0;
            break;
        }
    }   
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath 
{    
    switch (indexPath.section) {
        case CellCategoryIndexSimple: {
            switch (indexPath.row) {
                case SimpleCellIndexDefault: {
                    HLSTableViewCell *cell = [HLSTableViewCell cellForTableView:tableView];
                    cell.textLabel.text = NSLocalizedString(@"HLSTableViewCell", nil);
                    cell.selectionStyle = UITableViewCellSelectionStyleNone;
                    return cell;
                    break;
                }
                    
                case SimpleCellIndexValue1: {
                    HLSValue1TableViewCell *cell = [HLSValue1TableViewCell cellForTableView:tableView];
                    cell.textLabel.text = NSLocalizedString(@"HLSValue1TableViewCell", nil);
                    cell.detailTextLabel.text = NSLocalizedString(@"Details", nil);
                    cell.selectionStyle = UITableViewCellSelectionStyleNone;
                    return cell;
                    break;
                }
                    
                case SimpleCellIndexValue2: {
                    HLSValue2TableViewCell *cell = [HLSValue2TableViewCell cellForTableView:tableView];
                    cell.textLabel.text = NSLocalizedString(@"HLSValue2TableViewCell", nil);
                    cell.detailTextLabel.text = NSLocalizedString(@"Details", nil);
                    cell.selectionStyle = UITableViewCellSelectionStyleNone;
                    return cell;
                    break;
                }
                    
                case SimpleCellIndexSubtitle: {
                    HLSSubtitleTableViewCell *cell = [HLSSubtitleTableViewCell cellForTableView:tableView];
                    cell.textLabel.text = NSLocalizedString(@"HLSSubtitleTableViewCell", nil);
                    cell.detailTextLabel.text = NSLocalizedString(@"Details", nil);
                    cell.selectionStyle = UITableViewCellSelectionStyleNone;
                    return cell;                    
                    break;
                }
                    
                default: {
                    NSAssert(NO, @"Unsupported index");
                    return [UITableViewCell new];
                    break;
                }
            }
            break;
        }
            
        case CellCategoryIndexCustom: {
            switch (indexPath.row) {
                case CustomCellIndexXib: {
                    XibTableViewCell *cell = [XibTableViewCell cellForTableView:tableView];
                    cell.testLabel.text = NSLocalizedString(@"Custom cell from xib", nil);
                    cell.testImageView.image = [UIImage imageNamed:@"icn_bookmark.png"];
                    // Selection enabled to show that customisation works
                    return cell;
                    break;
                }
                    
                case CustomCellIndexProgrammatically:
                default: {
                    ProgrammaticTableViewCell *cell = [ProgrammaticTableViewCell cellForTableView:tableView];
                    cell.label.text = NSLocalizedString(@"Custom cell created programmatically", nil);
                    cell.selectionStyle = UITableViewCellSelectionStyleNone;
                    return cell;
                    break;
                }
            }
            break;
        }
            
        default: {
            NSAssert(NO, @"Unsupported index");
            return [UITableViewCell new];
            break;
        }
    }
}

#pragma mark UITableViewDelegate protocol implementation

- (UIView *)tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section
{
    HeaderView *headerView = [HeaderView view];
    switch (section) {
        case CellCategoryIndexSimple: {
            headerView.label.text = NSLocalizedString(@"Header: simple cells", nil);
            break;
        }
            
        case CellCategoryIndexCustom: {
            headerView.label.text = NSLocalizedString(@"Header: custom cells", nil);
            break;
        }
            
        default: {
            return nil;
            break;
        }
    }
    return headerView;
}

- (UIView *)tableView:(UITableView *)tableView viewForFooterInSection:(NSInteger)section
{
    FooterView *footerView = [FooterView view];
    footerView.label.text = NSLocalizedString(@"Section end", nil);
    return footerView;
}

- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath
{
    switch (indexPath.section) {
        case CellCategoryIndexSimple: {
            switch (indexPath.row) {
                case SimpleCellIndexDefault: {
                    return [HLSTableViewCell height];
                    break;
                }
                    
                case SimpleCellIndexValue1: {
                    return [HLSValue1TableViewCell height];
                    break;
                }
                    
                case SimpleCellIndexValue2: {
                    return [HLSValue2TableViewCell height];
                    break;
                }
                    
                case SimpleCellIndexSubtitle: {
                    return [HLSSubtitleTableViewCell height];
                    break;
                }
                    
                default: {
                    return 0;
                    break;
                }            
            }            
        }
            
        case CellCategoryIndexCustom: {
            switch (indexPath.row) {
                case CustomCellIndexXib: {
                    return [XibTableViewCell height];
                    break;
                }
                    
                case CustomCellIndexProgrammatically: {
                    return [ProgrammaticTableViewCell height];
                    break;
                }
                    
                default: {
                    return 0;
                    break;
                }            
            }
            break;
        }
            
        default: {
            return 0;
            break;
        }            
    }     
}

#pragma mark Localization

- (void)localize
{
    [super localize];
    
    self.title = NSLocalizedString(@"Table view cells", nil);
    [self.tableView reloadData];
}

@end
